(* 7) Bytecode interpreter *)
open Ast
open Bytecode
open Printf

let file = "output.ps";;

let execute_prog prog =
	let stack = Array.make 1024 0
	and globals = Array.make prog.num_globals 0 
	and pos = Array.make 4 0 (* 0 is pen, 1 is x, 2 is y, 3 is orientation *) 
	and oc = open_out file 
	
	in
	
	let rec exec fp sp pc = 
		if sp > 100 then raise (Failure("stack problem"));
		print_endline ("exec,fp="^string_of_int(fp)
		^",sp="^string_of_int(sp)
		^",pc="^string_of_int(pc)); match prog.text.(pc) with
			Lit i -> 
				print_endline("LIT_"^string_of_int i);
				stack.(sp) <- i; (* put the literat on the stack *) 
				exec fp (sp +1) (pc +1) (* increase the stack pointer and the program counter *)
		| Drp ->  
			print_endline("DRP");
			exec fp (sp -1) (pc +1) (* discard the top of the stack *)
		| Bin op ->  
			print_endline("BIN");
			let op1 = stack.(sp -2) 
			and op2 = stack.(sp -1) in
			stack.(sp - 2) <- 
			(let boolean i = if i then 1 else 0 in
					match op with
						Add -> op1 + op2
					| Sub -> op1 - op2
					| Mult -> op1 * op2
					| Div -> op1 / op2
					| Equal -> boolean (op1 = op2)
					| Neq -> boolean (op1 != op2)
					| Less -> boolean (op1 < op2)
					| Leq -> boolean (op1 <= op2)
					| Greater -> boolean (op1 > op2)
					| Geq -> boolean (op1 >= op2)) ;
				exec fp (sp - 1) (pc +1) (* stack is now 1 less since we don't need the top of it *)
				(* we add into where the first arg was, so now the second is dropped *)
		| Lod i ->  
			print_endline("LOD_"^string_of_int i);
			stack.(sp) <- globals.(i) ; 
			exec fp (sp +1) (pc +1) (* put a global with a given index on the stack, stack grows *)
		| Str i ->  
			print_endline("STR_"^string_of_int i);
			globals.(i) <- stack.(sp -1); 
			exec fp sp (pc +1) 
			(* store whatever is on top of the stack in a global, *)
			(* stack stays the same *)
		| Lfp i ->  
			print_endline("LFP_"^string_of_int i);
			stack.(sp) <- stack.(fp + i);
			exec fp (sp +1) (pc +1) (* load frame pointer relative on top of stack, stack grows *)
		| Sfp i ->  
			print_endline("SFP_"^string_of_int i);
			stack.(fp + i) <- stack.(sp -1); 
			exec fp sp (pc +1) (* put whatever is on top of the stack in fp relative location, *)
			(* stack stays the same *)

(* built-in function PRINT *)
		| Jsr(-1) ->  
			print_endline("PRINT");
			print_endline ("===========> print "^string_of_int stack.(sp -1));
			(* vv: was sp, not sp+1 *)
			exec fp sp (pc +1)
				
(* built-in function PENDOWN *)				
		| Jsr(-2) ->  
			print_endline("PENDOWN");
			pos.(0) <- 1; 
			print_endline ("pendown " ^(string_of_int stack.(sp -1)));
			exec fp sp (pc +1)
				
(* built-in function PENUP *)
		| Jsr(-3) ->  
			print_endline("PENUP");
			pos.(0) <- 0; 
			print_endline ("penup " ^ (string_of_int stack.(sp -1)));
			exec fp sp (pc +1)
				
(* built-in function FWD *)
		| Jsr(-4) ->   
			print_endline("FWD" ^ (string_of_int stack.(sp -1))); 
			
			(* TODO: improve this code, make it less imperative, use match or when *)
			if pos.(0) > 0 then print_endline ("=======> moveto " ^ (string_of_int pos.(1)) ^ " " ^ (string_of_int pos.(2)));
			if pos.(0) > 0 then fprintf oc "newpath\n%s\n" 
			("" ^ (string_of_int pos.(1)) ^ " " ^ (string_of_int pos.(2)) ^  " moveto");
			if pos.(3) = 0 then  pos.(1) <-  pos.(1) + stack.(sp -1);
			if pos.(3) = 90 then pos.(2) <-  pos.(2)+ stack.(sp -1);
		  if pos.(3) = 180 then pos.(1) <-  pos.(1) - stack.(sp -1);
			if pos.(3) = 270 then pos.(2) <-  pos.(2) - stack.(sp -1);
			if pos.(0) < 0 then raise (Failure ("drawing into negative x coord"));
			if pos.(1) < 0 then raise (Failure ("drawing into negative y coord"));
			if pos.(0) > 0 then print_endline ("===========> lineto " ^ (string_of_int pos.(1)) ^ " " ^ (string_of_int pos.(2)));
			if pos.(0) > 0 then fprintf oc "%s\nstroke\n" ("" ^ (string_of_int pos.(1)) ^ " " ^ (string_of_int pos.(2)) ^ " lineto");
			print_endline ("fwd " ^ string_of_int stack.(sp -1));
			exec fp sp (pc +1)
			
(* built-in function LEFT *)
		| Jsr(-5) ->  
			print_endline("LEFT"^ (string_of_int stack.(sp -1)));
			if stack.(sp -1) mod 90 != 0 then raise (Failure ("bad orientation, must be multiple of 90"));
			pos.(3) <- pos.(3) + 90; if (pos.(3) = 360) then pos.(3) <- 0; 
			print_endline ("left,orientation=" ^ string_of_int(pos.(3)) ^",top of stack is" ^ (string_of_int stack.(sp -1)));
			exec fp sp (pc +1)	
			
		| Jsr i ->  
			print_endline("JSR_"^string_of_int i);
			stack.(sp) <- pc + 1 ; 
			exec fp (sp +1) i
			
		| Ent i -> 
			print_endline("ENT_"^string_of_int i);
			stack.(sp) <- fp; 
			exec sp (sp + i +1) (pc +1)
		| Rts i ->  
			  print_endline("RTS_"^string_of_int i);
				let new_fp = stack.(fp) and new_pc = stack.(fp -1) in
				stack.(fp - i -1) <- stack.(sp -1);
				exec new_fp (fp - i) new_pc
		| Beq i ->  
			print_endline("BEQ_"^string_of_int i);
			exec fp (sp -1) (pc + if stack.(sp -1) = 0 then i else 1)
		| Bne i ->  
			print_endline("BNE_"^string_of_int i);
			exec fp (sp -1) (pc + if stack.(sp -1) != 0 then i else 1)
		| Bra i ->  
			print_endline("BRA_"^string_of_int i);
		  exec fp sp (pc + i)
		| Hlt ->  print_endline("HLT");()
	in
	let _ =   fprintf oc "%s\n\n2 setlinewidth\n" "%!PS"; print_endline("START");exec 0 0 0
	in print_endline("END"); fprintf oc "showpage\n\n%s\n" "%% end-of-output.ps"; close_out oc; 
